local mod = LibStub("AceAddon-3.0"):NewAddon("MarySue", "AceComm-3.0", "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("MarySue")
local ACD3 = LibStub("AceConfigDialog-3.0")
local LDB = LibStub:GetLibrary("LibDataBroker-1.1")

local updateFrame = CreateFrame("Frame")
local _G = _G.getfenv(0)
local strsplit = _G.strsplit
local strtrim = _G.strtrim
local tinsert = _G.tinsert
local tconcat = _G.table.concat
local UnitName = _G.UnitName
local IsInInstance = _G.IsInInstance
local GetRealmName = _G.GetRealmName
local UnitIsPlayer = _G.UnitIsPlayer
local UnitIsFriend = _G.UnitIsFriend
local GetTime = _G.GetTime
local _
local CreateFrame = _G.CreateFrame

local optFrame

local REQUEST = "REQ"
local PREFIX = "MARYSUE"
local THROTTLE_QUERY_TIME = 300
local CLIENT_THROTTLE_QUERY_TIME = 20

local bit_band = _G.bit.band
local bit_bor = _G.bit.bor
local bit_bxor = _G.bit.bxor
local type = _G.type
local pairs = _G.pairs
local wipe = _G.wipe
local math_max = _G.math.max
local tonumber = _G.tonumber
local string_sub = _G.string.sub

local db, hive, char
local queryTimes = {}
local display

local RP_LEVEL = {
	[1] = L["None"],
	[2] = L["Beginner"],
	[4] = L["Casual"],
	[8] = L["Full-time"],
	[16] = L["Mature (18+)"]
}

local CHARACTER_FLAGS = {
	[32] = L["Not roleplaying"],
	[64] = L["Out of Character"],
	[128] = L["In Character"],
	[256] = L["Looking for Contact"],
	[512] = L["Storyteller"]
}

local BEHAVIOR_VALUES = {
	[1] = L["Mouseover"],
	[2] = L["Target"]
}

local ldbObj = LDB:NewDataObject("MarySue", {
	type = "data source",
	icon = [[Interface\ICONS\Spell_Holy_Silence]],
	label = "MarySue",
	text  = "",
	
	OnClick = function(self, btn)
		if btn == "RightButton" then
			InterfaceOptionsFrame_OpenToCategory(optFrame)
		end
	end
})

local playerName = UnitName("player")
local function get(who)
	if who == playerName then
		return char
	else
		return hive[who]
	end
end

local function truncate(text, length)
	if #text - 3 > length then
		return string_sub(text, 1, length) .. "..."
	else
		return text
	end
end

local function getFullName(_, who)
	local t = who == nil and char or get(who)
	local ret
	if not t then
		ret = who
	else
		who = type(who) == "string" and who or UnitName("player")
		ret = strtrim(("%s %s %s"):format(t.prefix or "", who, t.suffix or ""))
	end
	return truncate(ret, db.behavior.truncateNameLength)
end

local getStatusString
do
	local level_buffer = {}
	local role_buffer = {}
	function getStatusString(flags)
		if not flags then return "", "", "" end
		
		wipe(level_buffer)
		wipe(role_buffer)
		for k, v in pairs(RP_LEVEL) do
			if bit_band(flags, k) == k then
				tinsert(level_buffer, v)
			end
		end

		for k, v in pairs(CHARACTER_FLAGS) do
			if bit_band(flags, k) == k then
				tinsert(role_buffer, v)
			end
		end
		
		local t1 = tconcat(level_buffer, ", ")
		local t2 = tconcat(role_buffer, ", ")
		wipe(role_buffer)
		if t1 and #t1 > 0 then tinsert(role_buffer, t1) end
		if t2 and #t2 > 0 then tinsert(role_buffer, t2) end
		local c = tconcat(role_buffer, " :: ")
		return c, t1 or "", t2 or ""
	end
end

local options = {
	type = "group",
	childGroups = "tab",
	args = {
		char = {
			type = "group",
			name = L["Character Profile"],
			args = {
				preview = {
					type = "header",
					name = getFullName,
					order = 1
				},
				prefix = {
					type = "input",
					name = L["Prefix"],
					desc = L["A prefix that shows before your name, like \"Lady\" or \"Crazy\""],
					get = function()
						return char.prefix
					end,
					set = function(info, v)
						char.prefix = v
					end,
					order = 10
				},
				suffix = {
					type = "input",
					name = L["Suffix"],
					desc = L["A suffix that shows after your name, like \"of Stormwind\" or \"Jinklestein\""],
					get = function()
						return char.suffix
					end,
					set = function(info, v)
						char.suffix = v
					end,
					order = 11
				},
				title = {
					type = "input",
					name = L["Title"],
					desc = L["Your formal title"],
					get = function()
						return char.title
					end,
					set = function(info, v)
						char.title = v
					end,
					order = 12,
					width = "full"
				},
				profile = {
					type = "input",
					name = L["Physical description"],
					desc = L["Describe your character's physical appearance, as others would perceive them."],
					multiline = 13,
					get = function()
						return char.profile
					end,
					set = function(info, v)
						char.profile = v
					end,
					width = "full"
				}
			}
		},
		flags = {
			type = "group",
			name = L["Flags & Status"],
			args = {
				level = {
					type = "multiselect",
					name = L["Roleplaying level"],
					get = function(info, key)
						return bit_band(char.flags, key) == key
					end,
					set = function(info, key, v)
						char.flags = bit_bxor(char.flags, key)
					end,
					values = RP_LEVEL
				},
				role = {
					type = "multiselect",
					name = L["Current role(s)"],
					get = function(info, key)
						return bit_band(char.flags, key) == key
					end,
					set = function(info, key, v)
						char.flags = bit_bxor(char.flags, key)
					end,
					values = CHARACTER_FLAGS
				}
			}
		},
		behavior = {
			type = "group",
			name = L["Behavior"],
			order = 105,
			args = {
				showOn = {
					type = "multiselect",
					name = L["Show on..."],
					get = function(info, key)
						return bit_band(db.behavior.showOn, key) == key
					end,
					set = function(info, key, val)
						db.behavior.showOn = bit_bxor(db.behavior.showOn, key)
					end,
					order = 1,
					values = BEHAVIOR_VALUES
				},
				disableInInstances = {
					type = "toggle",
					name = L["Disable in instances"],
					get = function()
						return db.behavior.disableInInstances
					end,
					set = function(info, v)
						db.behavior.disableInInstances = v
					end
				},
				sticky = {
					type = "toggle",
					name = L["Sticky Window"],
					desc = L["Keeps the description window open until you close it."],
					get = function()
						return db.behavior.sticky
					end,
					set = function(info, v)
						db.behavior.sticky = v
						mod:UpdateWindow()
					end
				},
				closeOnUntarget = {
					type = "toggle",
					name = L["Close on untarget"],
					get = function()
						return db.behavior.closeOnUntarget
					end,
					set = function(info, v)
						db.behavior.closeOnUntarget = v
					end,
					order = 150
				},
				max_height = {
					type = "range",
					name = L["Maximum profile height"],
					min = 100,
					max = 1000,
					step = 50,
					bigStep = 50,
					get = function()
						return db.behavior.max_display_height
					end,
					set = function(info, v)
						db.behavior.max_display_height = v
						mod:RefreshDisplay()
					end
				},
				tooltips = {
					type = "toggle",
					name = L["Tooltip info"],
					get = function()
						return db.behavior.useTooltips
					end,
					set = function(info, v)
						db.behavior.useTooltips = v
					end
				},
				truncate = {
					type = "range",
					name = L["Truncate long names to..."],
					min = 10,
					max = 100,
					step = 1,
					bigStep = 1,
					get = function()
						return db.behavior.truncateNameLength
					end,
					set = function(info, v)
						db.behavior.truncateNameLength = v
					end,
					order = 151
				}
			}
		}
	}
}

local defaults = {
	char = {
		suffix = "",
		prefix = "",
		title = "",
		profile = "",
		flags = 0
	},
	profile = {
		hive = {},
		behavior = {
			sticky = true,
			closeOnUntarget = true,
			max_display_height = 500,
			showOn = 3,
			disableInInstances = true,
			useTooltips = true,
			truncateNameLength = 50
		}
	}
}

function mod:OnInitialize()
	self:RegisterComm(PREFIX)
	self.db = LibStub("AceDB-3.0"):New("MarySueDB", defaults)
	db = self.db.profile
	char = self.db.char
	local realm = GetRealmName()
	db.hive[realm] = db.hive[realm] or {}
	hive = db.hive[realm]
	
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("MarySue", options)
	optFrame = ACD3:AddToBlizOptions("MarySue")
end

local function mouseoverDelay(f, t)
	mod:UpdateTooltip()
	f:SetScript("OnUpdate", nil)
end

function mod:OnEnable()
	self:Hook(GameTooltip, "SetUnit", "UPDATE_MOUSEOVER_UNIT", true)
    self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:UpdateWindow()
end

------------------------------------------------------------------------
--[[ These three hooks do the majority of the work ]]--
------------------------------------------------------------------------

function mod:OnTooltipSetUnit(frame)
	if not self:UPDATE_MOUSEOVER_UNIT() then return end
end

function mod:UPDATE_MOUSEOVER_UNIT(frame, unit)
	unit = unit or "mouseover"
	self.mouseoverUnit = unit	
	updateFrame:SetScript("OnUpdate", mouseoverDelay)
	
	if not self:Active() or bit_band(db.behavior.showOn, 1) ~= 1 then return end		
	self:UpdateDisplay(unit)
	return true
end

function mod:PLAYER_TARGET_CHANGED()
	if not self:Active() or bit_band(db.behavior.showOn, 2) ~= 2 then return end
	self:UpdateDisplay("target")
end

function mod:UpdateDisplay(unit)
	self:CloseDisplay()
	if UnitIsPlayer(unit) and UnitIsFriend("player", unit) then
		self:SendRequest(UnitName(unit))
	end
end

------------------------------------------------------------------------
------------------------------------------------------------------------


function mod:OnCommReceived( prefix, data, distribution, sender)
	if data == REQUEST then
		self:SendInfo(sender, distribution)
	else
		mod:ParseResponse(sender, strsplit("\1", data))
	end
end

function mod:ParseResponse(character, prefix, suffix, title, profile, flags)
	hive[character] = hive[character] or {}
	local t = hive[character]
	t.prefix = strtrim(prefix or "")
	t.suffix = strtrim(suffix or "")
	
	t.profile = strtrim(profile)
	t.profile = #t.profile == 0 and L["(No profile)"] or t.profile
	
	t.title = strtrim(title or "")
	t.flags = tonumber(strtrim(flags or "0"), 16)
	mod:Display(character, t)
	mod:UpdateTooltip()
end

function mod:SendRequest(to)
	local t = GetTime()
	local throttle = get(to) and CLIENT_THROTTLE_QUERY_TIME or THROTTLE_QUERY_TIME
	if t - (queryTimes[to] or 0) < throttle then
		if get(to) then
			mod:Display(to)
		end
		return
	end
	queryTimes[to] = t

	self:SendCommMessage(PREFIX, REQUEST, "WHISPER", to)
end
	
do
	local info = {}
	function mod:SendInfo(to, distribution)
		wipe(info)
		tinsert(info, char.prefix)
		tinsert(info, char.suffix)
		tinsert(info, char.title)
		tinsert(info, char.profile)
		tinsert(info, ("%x"):format(char.flags))
		local text = tconcat(info, "\1")
		self:SendCommMessage(PREFIX, text, distribution, distribution == "WHISPER" and to)
	end
end

function mod:RefreshDisplay()
	if display:IsVisible() then
		self:Display(self.lastWho, self.lastData, true)
	end
end

function mod:Display(who, data, force)	
	self.lastWho = who
	self.lastData = data
	if not who and not data then return end
	
	data = data or get(who)
	display:ClearLines()
	display.name:SetText(getFullName(nil, who))
	display.title:SetText(data.title or "")
	display.profile:SetText(data.profile)
	display.full_profile:SetText(data.profile)
	
	local status = getStatusString(data.flags)
	display.status:SetText(status == "" and L["(No status)"] or status)
	
	display:SetWidth(math_max(display.name:GetWidth(), display.title:GetWidth(), 300) + 45)
	display.scroll:SetWidth(display:GetWidth() - 40)
	
	display.profileFrame:SetWidth(display:GetWidth() - 40)
	display.profile:SetWidth(display.profileFrame:GetWidth())
	
	display.profileFrame:SetHeight(display.profile:GetHeight() + 20)
	
	local profileHeight, h = display.profile:GetHeight()
	if profileHeight > db.behavior.max_display_height then
		display.scroll:Show()
		display.scroll:SetHeight(db.behavior.max_display_height)
		display.full_profile:Hide()
		h = db.behavior.max_display_height
	else
		display.full_profile:Show()
		display.full_profile:SetWidth(display:GetWidth() - 20)
		display.scroll:Hide()
		h = display.full_profile:GetHeight()
	end
	
	display:SetHeight(display.name:GetHeight() + display.title:GetHeight() + display.status:GetHeight() + h + 40)
	display:Show()
end

function mod:CloseDisplay()
	if db.behavior.sticky and not db.behavior.closeOnUntarget then return end
	display:Hide()
end

do
	local left, right, colors = {}, {}, {}
	local function addLine(l, r, pos)
		if pos then
			tinsert(left, pos, l or "")
			tinsert(right, pos, r or "")
		else
			tinsert(left, l or "")
			tinsert(right, r or "")
		end
	end
	
	function mod:UpdateTooltip()
		if not self:Active() then return end
		if not db.behavior.useTooltips then return end
		
		local tt = _G.GameTooltip
		
		if not self.mouseoverUnit then return end
		local who = UnitName(self.mouseoverUnit)
		if not who then return end
		local data = get(who)
		if not data then return end
		
		wipe(colors)
		wipe(left)
		wipe(right)		
		
		for i = 1, tt:NumLines() do
			local l, r = _G["GameTooltipTextLeft" .. i], _G["GameTooltipTextRight" .. i]
			local lt, rt = l and l:GetText() or "", r and r:GetText() or ""

			tinsert(left, lt)
			tinsert(right, rt)
			colors[lt .. ":R"], colors[lt .. ":G"], colors[lt .. ":B"] = l:GetTextColor()
		end
		
		local color = (left[1] or ""):match("|%x%x%x%x%x%x%x%x%x")
		local full = getFullName(_, who)
		local tleft = left[1]
		left[1] = color and ("%s%s|r"):format(color, full) or full
		if tleft then
			colors[left[1] .. ":R"] = colors[tleft .. ":R"]
			colors[left[1] .. ":G"] = colors[tleft .. ":G"]
			colors[left[1] .. ":B"] = colors[tleft .. ":B"]
		end
		
		local idx = 2
		if data.title and data.title ~= "" then
			addLine(("|cffaaccff%s"):format(data.title), nil, idx)
			idx = idx + 1
		end
		
		local status, lstatus, rstatus = getStatusString(data.flags)
		if status and status ~= "" then
			addLine(lstatus, rstatus)
		end
		
		local last_l
		tt:ClearLines()
		for i = 1, #left do
			local l, r = left[i], right[i]
			local cr, cg, cb = colors[l .. ":R"], colors[l .. ":G"], colors[l .. ":B"]
			if r == "" and l ~= last_l then
				tt:AddLine(l, cr, cg, cb)
			elseif l ~= last_l then
				tt:AddDoubleLine(("|cffaaaacc%s"):format(l), ("|cffaaaacc%s"):format(r), cr, cg, cb)
			end
			last_l = l
		end
		
		tt:Show()
	end	
end

function mod:CreateDisplay()
	if display then return display end
	display = CreateFrame("Frame", "MarySueProfileFrame", UIParent)
	tinsert(UISpecialFrames, "MarySueProfileFrame")
	
	local f = display
	f:SetBackdrop({
		bgFile = [[Interface\DialogFrame\UI-DialogBox-Background]],
	    edgeFile = [[Interface\Tooltips\UI-Tooltip-Border]],
	    tile = true, tileSize = 16, edgeSize = 16,
	    insets = { left = 5, right = 5, top = 5, bottom = 5 }
	})
	f:SetWidth(150)
	f:SetHeight(150)
	f:SetPoint("LEFT")
	
	f:EnableMouse(true)
	f:SetMovable(true)
	f:SetClampedToScreen(true)	
	f:RegisterForDrag("LeftButton")
	f:SetScript("OnDragStart", function(frame)
		frame:StartMoving()
	end)
	f:SetScript("OnDragStop", function(frame)
		frame:StopMovingOrSizing()
	end)	
	
	f.name = f:CreateFontString(nil, nil, "GameFontNormalLarge")
	f.name:SetText("Name")
	f.name:SetPoint("TOPLEFT", display, "TOPLEFT", 10, -10)
	f.name:SetTextColor(0.5, 0.7, 1.0)
	
	f.title = f:CreateFontString(nil, nil, "GameFontNormal")
	f.title:SetText("Title")
	f.title:SetPoint("TOPLEFT", f.name, "BOTTOMLEFT", 0, -5)
	f.title:SetTextColor(0.4, 0.5, 1.0)

	f.status = f:CreateFontString(nil, nil, "GameFontNormal")
	f.status:SetText("Status")
	f.status:SetPoint("TOPLEFT", f.title, "BOTTOMLEFT", 0, -5)
	f.status:SetTextColor(1, 1, 0.8, 0.9)
	
	f.scroll = CreateFrame("ScrollFrame", "MarySueDescriptionScrollFrame", f, "UIPanelScrollFrameTemplate")
	f.scroll:SetHeight(200)
	f.scroll:SetWidth(400)
	f.scroll:SetPoint("TOPLEFT", f.status, "BOTTOMLEFT", 0, -5)

	f.profileFrame = CreateFrame("Frame", f.scroll)
	f.profileFrame:SetPoint("TOPLEFT")
	f.scroll:SetScrollChild(f.profileFrame)
	
	f.profile = f.profileFrame:CreateFontString(nil, nil, "GameFontNormal")
	f.profile:SetText("Profile")
	f.profile:SetJustifyV("TOP")
	f.profile:SetJustifyH("LEFT")
	f.profile:SetTextColor(1, 1, 1)
	f.profile:SetPoint("TOPLEFT")
	
	f.full_profile = f:CreateFontString(nil, nil, "GameFontNormal")
	f.full_profile:SetText("Profile")
	f.full_profile:SetJustifyV("TOP")
	f.full_profile:SetJustifyH("LEFT")
	f.full_profile:SetTextColor(1, 1, 1)
	f.full_profile:SetPoint("TOPLEFT", f.status, "BOTTOMLEFT", 0, -5)
	
	f.close = CreateFrame("Button", nil, f, "UIPanelCloseButton")
	f.close:SetPoint("TOPRIGHT")
	
	f.ClearLines = function(self)
		self.name:SetText("")
		self.title:SetText("")
		self.status:SetText("")
		self.profile:SetText("")
	end
	
	return display
end

function mod:UpdateWindow()
	if db.behavior.sticky then
		display.close:Show()
	else
		display.close:Hide()
	end
end

function mod:Active()
	if db.behavior.disableInInstances then		
		return IsInInstance() ~= 1
	end
	return true	
end

mod:CreateDisplay()